//	Draw4DApp.swift
//
//	© 2023 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


@main
struct Draw4DApp: App {
	
	init() {

		if !UserDefaults.standard.bool(forKey: "legacy files have been converted") {
			ConvertLegacyFiles()
		}

#if targetEnvironment(simulator)
		//	When creating screenshots in the simulator,
		//	we'll need to know where to put source files
		//	for the desired drawings.
		PrintDocumentDirectoryPathToConsole()
#endif
	}
	
	var body: some Scene {
	
		DocumentGroup(newDocument: { Draw4DDocument() }) { file in

			Draw4DContentView(
				document: file.document,
				fileURL: file.fileURL
			)
		}
	}
}

func ConvertLegacyFiles() {

	do {
		let theFileManager = FileManager.default
		
		//	Get a reference to the app's Documents directory.
		let theDocumentsDirectory = try theFileManager.url(
				for: .documentDirectory,
				in: .userDomainMask,
				appropriateFor: nil,
				create: true)
	
		//	What does the Documents directory contain?
		let theDirectoryContents = try theFileManager.contentsOfDirectory(
			at: theDocumentsDirectory,
			includingPropertiesForKeys: [.isRegularFileKey, .nameKey],
			options: [.skipsHiddenFiles])

		//	Consider each file...
		for theFileURL in theDirectoryContents {
			
			do {
				
				let theResourceValues = try theFileURL.resourceValues(forKeys: [.nameKey, .isRegularFileKey])
				if !(theResourceValues.isRegularFile ?? false) {
					continue
				}
				guard let theFileName = theResourceValues.name else {
					continue
				}

				//	If this file is a drawing file,
				//	change its filename extension from ".txt" to ".draw4d" .
				//
				//		Note:  In the unlikely event that the user has placed
				//		some other .txt file (not a 4D Draw drawing) into the directory,
				//		Draw4DThumbnailProvider's provideThumbnail() method
				//		will notice that decodeFileFormat() fails, and will call
				//
				//			handler(nil, Draw4DThumbnailError.invalidDrawingData)
				//
				//		to report the problem to the document picker.
				//		The document picker thus knows to keep that file disabled,
				//		and not let the user open it in 4D Draw.
				//
				if theFileName.hasSuffix(".txt") {
					let theNewFileURL = theFileURL.deletingPathExtension().appendingPathExtension("draw4D")
					try theFileManager.moveItem(at: theFileURL, to: theNewFileURL)
				}
				
				//	If this file is the drawings manifest or a thumbnail image,
				//	delete it.
				if (theFileName == "4DDrawDrawingsManifest.xml"
				 || theFileName.hasSuffix(".png")
				) {
					try theFileManager.removeItem(at: theFileURL)
				}

			} catch {
				//	In the unlikely event that we failed to delete/rename a file,
				//	keep going in hopes of deleting/renaming the other files.
			}
		}
		
		//	The legacy files have now all been renamed/deleted.
		//	For this user on this device, there's no need
		//	to call ConvertLegacyFiles() ever again.
		UserDefaults.standard.set(true, forKey:"legacy files have been converted")

	} catch {
		
		//	In the unlikely case that something went wrong,
		//	let the user push onward without converting the files.
		//	We'll try the conversion again the next time the user launches the app.
	}
}
