//	Draw4DContentView.swift
//
//	© 2023 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI
import simd


enum PanelType {
	case noPanel
	case optionsPanel
	case rotationPanel
	case exportPanel
	case helpHowToDrawPanel
	case helpIdeasPanel
	case contactPanel
	case translatorsPanel
}


struct Draw4DContentView: View {

	@Environment(\.undoManager) var undoManager

	@ObservedObject var document: Draw4DDocument
	let fileURL: URL?

	@State var itsActivePanel: PanelType = .noPanel
	
	//	A renderer is a potentially expensive resource
	//	(it may have to load textures, etc.), so let's
	//	let the GeometryGamesView and the GeometryGamesExportView
	//	share a single renderer, instead of forcing each
	//	to create its own.
	//
	//		Note #1  The renderer's init() may fail,
	//		and @StateObject doesn't seem willing to hold an optional
	//		(because an optional doesn't conform to ObservableObject),
	//		so I've wrapped the Draw4DRenderer
	//		in an OptionalDraw4DRenderer as a workaround.
	//
	//		Note #2  itsOptionalRenderer isn't really an observed state,
	//		it's just a resource that's shared among our subviews.
	//		(But it's not shared between Scenes -- each Scene
	//		gets its own OptionalDraw4DRenderer instance.)
	//
	class OptionalDraw4DRenderer: ObservableObject {
		
		var itsRenderer: Draw4DRenderer?
		init(_ renderer: Draw4DRenderer?) {
			itsRenderer = renderer
		}
	}
	@StateObject var itsOptionalRenderer
		= OptionalDraw4DRenderer(Draw4DRenderer())
	
	//	State variables for use in the gesture handlers.
	@State var dragPurpose: DragPurpose = .unknown
	@State var recentPoints: [RecentPoint] = []
	@State var previousAngle: RecentAngle? = nil


	var body: some View {
		
		let theTranslatorThanks = [
			GeometryGamesTranslatorThanks(pre: "grazie a", name: "Daniela, Isabella Eva e Davide Bigatti", post: nil, lang: "it"),
			GeometryGamesTranslatorThanks(pre: "obrigado a", name: "Nuno Fernandes", post: nil, lang: "pt")
		]

		if gMakeScreenshots {

			setOrientationForScreenshot(
				modelData: document,
				fileURL: fileURL)
		}
	
		//	When I someday require iOS 17.1 or higher,
		//	I may remove the following workaround
		//	for the iOS 17.0.x SwiftUI document-based app nav bar bug.
		let theIOSVersion = ProcessInfo.processInfo.operatingSystemVersion
		let theSwiftUIBugIsPresent = (
			theIOSVersion.majorVersion == 17
		 && theIOSVersion.minorVersion == 0 )
		
		//	This top-level ZStack lets an optional panel overlay the main view.
		return ZStack(alignment: .bottom) {

			//	The system-provided UndoManager should always
			//	be present when body is called, but I couldn't
			//	find any documentation promising this.
			//	So let's test whether it's non-nil,
			//	and display an error message if it's not.
			//
			//		Note:  An alternative strategy could be
			//		to let each View access @Environment(\.undoManager)
			//		on its own.  But while all Views contained
			//		in the main content view would get the same UndoManager,
			//		Views contained in a popover would get
			//		a different UndoManager.  That's why we
			//		explicitly pass theUndoManager down through
			//		the view hierarchy, to ensure that all Views,
			//		even those in a popover, are using the same
			//		UndoManager.
			//
			if let theUndoManager = undoManager {
	
				//	Main animation view
				if let theRenderer = itsOptionalRenderer.itsRenderer {

					GeometryReader() { geometryReaderProxy in
					
						GeometryGamesViewRep(
							modelData: document,
							renderer: theRenderer
						)
						//
						//	The order in which the gestures get attached is important:
						//
						//		.gesture(DragGesture()…)
						//		.gesture(RotationGesture()…)
						//
						//	works fine, but
						//
						//		.gesture(RotationGesture()…)
						//		.gesture(DragGesture()…)
						//
						//	fails to properly handle the DragGesture.
						//	I can't find any documentation about this
						//	(nor even any blog posts or other discussion online),
						//	but I'm _guessing_ that the explanation for this behavior
						//	is the following:  In the first case shown above,
						//	as soon as the user places two fingers on the display,
						//	SwiftUI knows the DragGesture has failed and moves on
						//	to consider the RotationGesture.  But in the second case,
						//	when the user drags a finger on the display, SwiftUI
						//	doesn't give up hope for the RotationGesture,
						//	so the DragGesture receives no onChanged messages
						//	(even though it does eventually receive a final onEnded message).
						//
						.gesture(draw4DDragGesture(
									modelData: document,
									viewSize: geometryReaderProxy.size,
									undoManager: theUndoManager,
									dragPurpose: $dragPurpose,
									recentPoints: $recentPoints))
						.gesture(draw4DRotationGesture(
									modelData: document,
									previousAngle: $previousAngle,
									dragPurpose: $dragPurpose))	//	needed only for workaround
						.simultaneousGesture(
							//	If we made a TapGesture the top-priority gesture,
							//	the user would have to wait for it to fail before
							//	getting a response to a DragGesture.  So let's test
							//	for a TapGesture in parallel with the Drag/Rotate
							//	gestures, and use it only to dismiss an open panel.
							TapGesture()
							.onEnded() { _ in itsActivePanel = .noPanel })
					}
					.edgesIgnoringSafeArea(.bottom)	//	set to .all to underlap nav bar as well
				}
				
				//	Controls
				VStack() {

					ZStack() {

						//	Controls along the top edge
						//
						//	Use a ZStack so that the Undo/Redo buttons may be centered.
						//
						ZStack {

							HStack(alignment: .top) {
								
								Spacer()

								UndoButton(
									modelData: document,
									undoManager: theUndoManager
								)
								.disabled(!document.itCanUndo)
								
								RedoButton(
									modelData: document,
									undoManager: theUndoManager
								)
								.disabled(!document.itCanRedo)

								Spacer()
							}

							HStack(alignment: .top) {
								
								Spacer()
								
								LockButton(
									modelData: document,
									sourceDocumentURL: fileURL,
									undoManager: theUndoManager
								)
							}
						}
						.frame(maxHeight: .infinity, alignment: .top)

						Group() {
						
							//	Note:  Some of the panels get leading-aligned, others trailing-aligned.
							switch itsActivePanel {
							
							case .noPanel:
								Spacer()
								
							case .optionsPanel:
								OptionsView(modelData: document)
								.frame(maxWidth: .infinity, alignment: .leading)
								
							case .rotationPanel:
								RotationView(modelData: document, undoManager: theUndoManager)
								.frame(maxWidth: .infinity, alignment: .leading)
								
							case .exportPanel:
								let theDrawingName = fileURL?.deletingPathExtension().lastPathComponent
								if let theRenderer = itsOptionalRenderer.itsRenderer {
									GeometryGamesExportView(
										modelData: document,
										exportRenderer: theRenderer,
										folderName: theDrawingName,
										imageName: theDrawingName)
									.preferredColorScheme(.light)	//	preferredColorScheme() works
										//	in 4D Maze, but not in 4D Draw, possibly because when
										//	SwiftUI translates everyting to UIKit under the hood,
										//	it handles the "presentation" contexts differently
										//	in this document-based app.  As a workaround,
										//	the info.plist file sets Appearance = Light,
										//	with the drawback that the Nav Bar is also light.
									.frame(maxWidth: .infinity, alignment: .trailing)
								}
								
							case .helpHowToDrawPanel:
								Draw4DHelpHowView()
								.frame(maxWidth: .infinity, alignment: .trailing)
								
							case .helpIdeasPanel:
								Draw4DHelpIdeasView()
								.frame(maxWidth: .infinity, alignment: .trailing)
								
							case .contactPanel:
								GeometryGamesContactView()
								.preferredColorScheme(.light)	//	preferredColorScheme() works
										//	in 4D Maze, but not in 4D Draw, possibly because when
										//	SwiftUI translates everyting to UIKit under the hood,
										//	it handles the "presentation" contexts differently
										//	in this document-based app.  As a workaround,
										//	the info.plist file sets Appearance = Light,
										//	with the drawback that the Nav Bar is also light.
								.frame(maxWidth: .infinity, alignment: .trailing)
								
							case .translatorsPanel:
								GeometryGamesTranslatorsView(theTranslatorThanks)
								.frame(maxWidth: .infinity, alignment: .trailing)
							}
						}
						.padding(EdgeInsets(
									top: geometryGamesPanelMargin,
									leading: geometryGamesPanelMargin,
									bottom: 0.0,	//	OK to sit close to buttons
									trailing: geometryGamesPanelMargin))
						.frame(maxHeight: .infinity, alignment: .bottom)
					}
					
					//	Controls along the bottom edge
					HStack(alignment: .bottom) {

						//	The buttons already have their own tappable padding,
						//	so the AdaptiveStack doesn't need to provide any extra spacing.
						AdaptiveStack(spacing: 0) {
							OptionsButton(activePanel: $itsActivePanel)
							RotationButton(activePanel: $itsActivePanel)
								.disabled(!document.itCanRotateQuarterTurns)
						}

						Spacer()

						TouchModePicker(
							touchMode: $document.itsTouchMode,
							canMovePoint: $document.itCanMovePoint,
							canAddPoint: $document.itCanAddPoint,
							canDeletePoint: $document.itCanDeletePoint,
							canAddEdge: $document.itCanAddEdge,
							canDeleteEdge: $document.itCanDeleteEdge
						)
						
						Spacer()

						//	The buttons already have their own tappable padding,
						//	so the AdaptiveStack doesn't need to provide any extra spacing.
						AdaptiveStack(spacing: 0) {
							ExportButton(activePanel: $itsActivePanel)
							HelpButton(activePanel: $itsActivePanel)
						}
					}
				}
				
			} else {
			
				Color.yellow
				
				VStack() {
				
					Text("The system-provided UndoManager is unexpectedly missing.  Please let Jeff Weeks know this happened, via the address given at")
					.multilineTextAlignment(.center)
					.foregroundColor(.black)
					.frame(maxWidth: 360.0, alignment: .center)
				
					if let theContactURL = URL(string: "http://www.geometrygames.org/contact.html") {
						Link(
						"www.geometrygames.org/contact.html",
						 destination: theContactURL)
					} else {
						Text("www.geometrygames.org/contact.html")
					}
				}
				.frame(maxHeight: .infinity, alignment: .center)
			}
			
			if theSwiftUIBugIsPresent {
			
				VStack(spacing: 8.0) {
					
					Text("iOS 17.1 required")
					.font(.title)
					.foregroundColor(.red)

					Text("4D Draw runs correctly on iOS 15, 16 and 17.1 or later. But it won't run correctly on iOS 17.0.x because a bug in iOS 17.0.x fails to show the Back button and the drawing title in SwiftUI apps. Please update your device to iOS 17.1 or later to solve this problem.")
					.font(.title2)
					.foregroundColor(.black)
				}
				.padding(24.0)
				.background(.white)
				.cornerRadius(16.0)
				.overlay(
					RoundedRectangle(cornerRadius: 16.0)
					.stroke(.black, lineWidth: 4.0)
				)
				.padding(32.0)
				.frame(maxWidth: .infinity, maxHeight: .infinity, alignment: .center)
			}
		}
		.onAppear() {

			document.itsModelChangeCount += 1
			
			if gMakeScreenshots {
				itsActivePanel = setPanelForScreenshot(fileURL: fileURL)
			}
		}
		//	When most users are running iOS 17 or later
		//	we may eliminate DrawingTitleStyle entirely
		//	(assuming no more surprises in iOS 18 or later).
		.modifier(DrawingTitleStyle())
	}

	struct DrawingTitleStyle: ViewModifier {
	
		func body(content: Content) -> some View {
	
			if #available(iOS 17, *) {
			
				//	On iOS 17, the title is displayed .inline by default.
				content
			
			} else if #available(iOS 16, *) {

				//	On iOS 16, the title is displayed .large by default.
				//	Override that default to display an .inline title.
				content
				.navigationBarTitleDisplayMode(.inline)

			} else {

				//	On iOS 15, navigationBarTitleDisplayMode(.inline)
				//	fails to show the title at all (!).  Fortunately
				//	if we omit navigationBarTitleDisplayMode(),
				//	then the title shows as desired.
				content
			}
		}
	}
}


struct UndoButton: View {

	let modelData: Draw4DDocument
	let undoManager: UndoManager

	var body: some View {

		Button {
		
			undoManager.undo()
			
			modelData.updateCapabilityFlags(undoManager: undoManager)

		} label: {
			Image(systemName: "arrowshape.turn.up.backward.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}

struct RedoButton: View {

	let modelData: Draw4DDocument
	let undoManager: UndoManager

	var body: some View {

		Button {
		
			undoManager.redo()
			
			modelData.updateCapabilityFlags(undoManager: undoManager)

		} label: {
			Image(systemName: "arrowshape.turn.up.forward.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


struct LockButton: View {
	
	@ObservedObject var modelData: Draw4DDocument
	let sourceDocumentURL: URL?
	let undoManager: UndoManager

	var body: some View {

		let theFileName = String(sourceDocumentURL?.deletingPathExtension().lastPathComponent ?? "")

		return Menu() {

			if modelData.itsContentIsLocked {

				Button() {

					modelData.itsContentIsLocked = false

					modelData.updateCapabilityFlags(undoManager: undoManager)

				} label: {
					Label("UnlockFigure \(theFileName)", systemImage: "lock.open.fill")
				}

			} else {

				Button() {

					modelData.itsContentIsLocked = true

					modelData.updateCapabilityFlags(undoManager: undoManager)

				} label: {
					Label("LockFigure \(theFileName)", systemImage: "lock.fill")
				}
			}

		} label: {

			Image(systemName: (modelData.itsContentIsLocked ? "lock.fill" : "lock.open.fill"))
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


struct OptionsButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .optionsPanel
		} label: {
			Image(systemName: "switch.2")	//	or "slider.horizontal.3"
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}

struct OptionsView: View {

	@ObservedObject var modelData: Draw4DDocument

	var body: some View {

		VStack() {

			Toggle(isOn: $modelData.itsBoxIsEnabled) {
				Text("Box")
			}

			Toggle(isOn: $modelData.itsSnapToGridIsEnabled) {
				Text("SnapToGrid")
			}

			Toggle(isOn: $modelData.itsInertiaIsEnabled) {
				Text("Inertia")
			}
		}
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		//	preferredColorScheme() works in 4D Maze, but not in 4D Draw,
		//	possibly because when SwiftUI translates everyting to UIKit
		//	under the hood, it handles the "presentation" context differently.
		//	As a workaround, the info.plist file sets Appearance = Light,
		//	with the drawback that the Nav Bar is also light.
		.background(Color(UIColor.systemBackground))
		.preferredColorScheme(.light)
		.cornerRadius(geometryGamesCornerRadius)
	}
}


struct RotationButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .rotationPanel
		} label: {
			Image(systemName: "arrow.triangle.2.circlepath.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}

struct RotationView: View {

	@ObservedObject var modelData: Draw4DDocument
	let undoManager: UndoManager
	
	let theAnimationDurationSlow = 4.0	//	seconds
	let theAnimationDurationFast = 2.0	//	seconds

	var body: some View {

		VStack(spacing: 16.0) {

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationFast,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(0.0, 0.0, +1.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(0.0, 0.0, -1.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("X → Y")") }

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationFast,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(+1.0, 0.0, 0.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(-1.0, 0.0, 0.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("Y → Z")") }

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationFast,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(0.0, +1.0, 0.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(0.0, -1.0, 0.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("Z → X")") }

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationSlow,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(+1.0, 0.0, 0.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(+1.0, 0.0, 0.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("W → X")") }

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationSlow,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(0.0, +1.0, 0.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(0.0, +1.0, 0.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("W → Y")") }

			Button {
				modelData.itsAnimatedRotation = AnimatedRotation(
					itsTotalDuration: theAnimationDurationSlow,
					itsTotalAngle: 0.5 * Double.pi,
					itsStartTime: CFAbsoluteTimeGetCurrent(),
					itsLeftImaginaryUnitVector:  SIMD3<Double>(0.0, 0.0, +1.0),
					itsRightImaginaryUnitVector: SIMD3<Double>(0.0, 0.0, +1.0),
					itsUndoManager: undoManager)
				
			} label: { Text("Rotate \("W → Z")") }
		}
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		//	preferredColorScheme() works in 4D Maze, but not in 4D Draw,
		//	possibly because when SwiftUI translates everyting to UIKit
		//	under the hood, it handles the "presentation" context differently.
		//	As a workaround, the info.plist file sets Appearance = Light,
		//	with the drawback that the Nav Bar is also light.
		.background(Color(UIColor.systemBackground))
		.preferredColorScheme(.light)
		.cornerRadius(geometryGamesCornerRadius)
		.disabled(!modelData.itCanRotateQuarterTurns)
	}
}


struct TouchModePicker: View {

	@Binding var touchMode: TouchMode
	@Binding var canMovePoint: Bool
	@Binding var canAddPoint: Bool
	@Binding var canDeletePoint: Bool
	@Binding var canAddEdge: Bool
	@Binding var canDeleteEdge: Bool

	var body: some View {

		//	This code uses UIKit methods to style the underlying UISegmentedControl,
		//	which makes me uncomfortable.  Let's hope some future version of SwiftUI
		//	provides a cleaner way to do it.
		//
		//		Note:  These settings would apply to any and all
		//		UISegmentedControls that we create, although
		//		in practice 4D Draw creates only this one.
		//
		UISegmentedControl.appearance().selectedSegmentTintColor
			= UIColor.init(displayP3Red: 0.00, green: 1.00, blue: 0.75, alpha: 1.0)
				//	iOS will desaturate that color
		UISegmentedControl.appearance().backgroundColor = UIColor.white
				//	iOS will replace that pure white with a light grey
		UISegmentedControl.appearance().setTitleTextAttributes(
			[.foregroundColor: UIColor(displayP3Red: 0.0, green: 0.0, blue: 1.0, alpha: 1.0)],
			for: .selected)
		UISegmentedControl.appearance().setTitleTextAttributes(
			[.foregroundColor: UIColor(Color.accentColor)],
			for: .normal)

		//	Create a segmented Picker.
		return Picker("foo", selection: $touchMode) {

			//	Note #1
			//	In principle we could use a loop to create the Picker's items
			//
			//		ForEach(TouchMode.allCases, id: \.self) {theTouchMode in
			//			Image(theTouchMode.imageName)
			//				.renderingMode(.template)
			//		}
			//
			//	but that would seem to make it more complicated
			//	to selectively enable/disable the items.
			//	So let's stick with manually listing the items.
			
			//	Note #2
			//	Textual labels like
			//
			//		Text(theTouchMode.nameKey)
			//
			//	look great, but alas there's not enough space for them all,
			//	not even on an iPad.  If some future version of SwiftUI
			//	supports the equivalent of UIKit's setApportionsSegmentWidthsByContent:NO,
			//	we could consider using textual labels in a "regular size class"
			//	and image labels in a "compact size class".

			//	Note #3
			//	The modifier .renderingMode(.template) asks SwiftUI to replace
			//	the image's colors with the accent color, which is exactly what
			//	we want.  (The red and green colors in the raw images would be
			//	a bit gaudy in this version of 4D Draw, and in any case
			//	invite confusion with the interpretation of any red or green edges
			//	in the drawing itself.)

			//	Neutral mode is always enabled
			Image(TouchMode.neutral.imageName)
				.renderingMode(.template)
				.tag(TouchMode.neutral)

			if canMovePoint {
				Image(TouchMode.movePoints.imageName)
					.renderingMode(.template)
					.tag(TouchMode.movePoints)
			}

			if canAddPoint {
				Image(TouchMode.addPoints.imageName)
					.renderingMode(.template)
					.tag(TouchMode.addPoints)
			}

			if canDeletePoint {
				Image(TouchMode.deletePoints.imageName)
					.renderingMode(.template)
					.tag(TouchMode.deletePoints)
			}

			if canAddEdge {
				Image(TouchMode.addEdges.imageName)
					.renderingMode(.template)
					.tag(TouchMode.addEdges)
			}

			if canDeleteEdge {
				Image(TouchMode.deleteEdges.imageName)
					.renderingMode(.template)
					.tag(TouchMode.deleteEdges)
			}
		}
		.pickerStyle(SegmentedPickerStyle())
		.fixedSize()
		.padding(geometryGamesControlPadding)	//	for spacing only, not to expand tappable area
	}
}


struct ExportButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = .exportPanel
		} label: {
			Image(systemName: "square.and.arrow.up")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


struct HelpButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Menu() {

			Button() {

				activePanel = .helpHowToDrawPanel

			} label: {
				Label("How to draw", systemImage: "questionmark.circle")
			}
		
			Button() {

				activePanel = .helpIdeasPanel

			} label: {
				Label("Construction ideas", systemImage: "lightbulb")
			}
		
			Button() {

				activePanel = .contactPanel
				
			} label: {
				GeometryGamesHelpMenuContactItemLabel()
			}
			
			Button() {

				activePanel = .translatorsPanel
				
			} label: {
				GeometryGamesHelpMenuTranslatorsItemLabel()
			}
			
		} label: {

			Image(systemName: "questionmark.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


func setOrientationForScreenshot(
	modelData: Draw4DDocument,
	fileURL: URL?
) {

	let theFileName = String(fileURL?.deletingPathExtension().lastPathComponent ?? "")

	let theOrientation: simd_quatd
	switch theFileName {
	
	case "vertex-first hypercube (step 3)":
		theOrientation = simd_quatd(
			real: 0.957772,
			imag: SIMD3<Double>(-0.250343,  0.030159, -0.138172))

	case "vertex-first hypercube (step 1)":
		theOrientation = simd_quatd(
			real: 0.957772,
			imag: SIMD3<Double>(-0.250343,  0.030159, -0.138172))

	case "alternating vertices of hypercube":
		theOrientation = simd_quatd(
			real: 0.952200,
			imag: SIMD3<Double>(-0.223655, -0.203244,  0.044552))

	case "hypertetrahedron":
		theOrientation = simd_quatd(
			real: 0.966959,
			imag: SIMD3<Double>(-0.254932,  0.000000,  0.000000))

	default:
		theOrientation = gQuaternionIdentity

	}

	modelData.itsOrientation = theOrientation
}

func setPanelForScreenshot(
	fileURL: URL?
) -> PanelType {

	let theFileName = String(fileURL?.deletingPathExtension().lastPathComponent ?? "")

	let thePanelType: PanelType
	switch theFileName {
	
	case "alternating vertices of hypercube":
		thePanelType = .helpHowToDrawPanel

	case "hypertetrahedron":
		thePanelType = .helpIdeasPanel

	default:
		thePanelType = .noPanel
	}

	return thePanelType
}


//struct Draw4DContentView_Previews: PreviewProvider {
//	static var previews: some View {
//		Draw4DContentView(document: .constant(Draw4DDocument()))
//	}
//}
